from .constants import LAUNCH_TIME
from .exceptions import ConfigNotFoundError, InvalidConfigError
from .file_handling import FileManager
from .utils import get_monthrange
from .validator import validate

INVALID_OFFSET_MESSAGE = (
    "Invalid offset value: {value}\n(value must be an integer greater than 2 and less than 250)"
)
INVALID_YEARS_MESSAGE = "Invalid years in release date: {value}"
INVALID_TITLES_MESSAGE = "Invalid titles: {value}"
INVALID_PROGRESS_BAR_MESSAGE = "Invalid progress bar value: {value}\n(value must be a boolean)"


class TelegraphParserConfig:
    def __init__(self, config_file_path: str) -> None:
        self.config_file_path = config_file_path
        self.__load_config()
        try:
            self.parse_config()
        except (KeyError, TypeError) as error:
            raise InvalidConfigError(error)
        self.__calculate_totals()

    def __load_config(self) -> None:
        self.config = FileManager.load_yaml(self.config_file_path)
        if self.config is None:
            raise ConfigNotFoundError(path=self.config_file_path)

    def parse_config(self) -> None:
        """Parses the configuration YAML file

        Your own config-parser should be structured as follows:
            self.{NAME} = self.config.get("{CONFIGURATION_PARAMETER_NAME}")

        Example:
            self.offset = self.config.get("offset")

        NOTE: if you want to validate the parameter value, use the validate() function
        """
        self.offset = validate(
            self.config.get("offset"),
            "integer",
            default_value=1,
            value_range=(2, 250),
            exception_message=INVALID_OFFSET_MESSAGE,
        )
        self.release_date = validate(
            self.config.get("release_date"),
            "integer_list",
            value_range=(0, LAUNCH_TIME.year),
            exception_message=INVALID_YEARS_MESSAGE,
        )
        self.titles = validate(
            self.config.get("titles"),
            "any_list_wn",
            exception_message=INVALID_TITLES_MESSAGE,
        )
        self.progress_bar = validate(
            self.config.get("progress_bar"),
            "boolean",
            exception_message=INVALID_PROGRESS_BAR_MESSAGE,
        )

    def __calculate_totals(self) -> None:
        self.total_months = LAUNCH_TIME.month if self.release_date == [LAUNCH_TIME.year] else 12
        self.total_days = (
            sum(get_monthrange(month) for month in range(1, self.total_months + 1))
            if self.total_months != 12
            else 366
        )
        self.total_urls = len(self.titles) * self.offset * self.total_days



import os

import yaml

YAML_EXTENSIONS = (".yml", ".yaml")


class FileManager:
    @staticmethod
    def create_folder(folder_path: str) -> None:
        os.makedirs(folder_path, exist_ok=True)

    @staticmethod
    def join_paths(*paths: str) -> str:
        return os.path.join(*paths)

    @staticmethod
    def dump_binary_data(file_path: str, data: bytes) -> None:
        with open(file_path, "wb") as file:
            file.write(data)

    @staticmethod
    def load_yaml(file_path: str) -> dict[any, any] | None:
        """Loads a YAML file and returns a dictionary

        :param file_path: (String) YAML file path without file extension

        :return: dictionary if successful, None otherwise
        """
        for extension in YAML_EXTENSIONS:
            try:
                with open(file_path + extension, encoding="utf-8") as file:
                    return yaml.safe_load(file)
            except FileNotFoundError:
                ...

    @staticmethod
    def dump_yaml(file_path: str, data: dict[any, any]) -> None:
        """Dump a dictionary into a YAML file

        :param file_path: (String) YAML file path
        :param data: dictionary to dump
        """
        with open(file_path, "w", encoding="utf-8") as file:
            yaml.dump(data, file)



from abc import ABC, abstractmethod


class ApplicationException(ABC, Exception):
    @staticmethod
    @abstractmethod
    def get_error_message(exception: "ApplicationException") -> None: ...

    @property
    @abstractmethod
    def message(self) -> str: ...



from .exceptions import InvalidValidationTypeError, InvalidValidationValueError


def check_condition(condition: bool, value: any, exception_message: str) -> None:
    if not condition:
        raise InvalidValidationValueError(value, exception_message)


def check_range(value: int, value_range: tuple[int, int] | None) -> bool:
    # Checks that the integer value is within the range
    # if value_range is specified
    return value_range is None or value_range[0] <= value <= value_range[1]


def boolean(value: any, exception_message: str) -> bool:
    # Checks if the value is of type bool
    check_condition(isinstance(value, bool), value, exception_message)

    return value


def integer(value: int, exception_message: str, value_range: tuple[int, int] | None) -> int:
    # Checks if the value is of type int,
    # and if the range of acceptable values is specified,
    # verifies that the value is within the range
    check_condition(
        isinstance(value, int) and check_range(value, value_range),
        value,
        exception_message,
    )

    return value


def integer_list(
    values: any, exception_message: str, value_range: tuple[int, int] | None
) -> list[int]:
    # Checks if the value is of type list,
    # and ensures that all values in the list are of type int.
    # If the range of acceptable values is specified,
    # verifies that each value in the list is within the range
    check_condition(
        isinstance(values, list)
        and all(isinstance(value, int) and check_range(value, value_range) for value in values),
        values,
        exception_message,
    )

    return values


def any_list_wn(values: any, exception_message: str) -> list[any]:
    # Checks if the value is of type list,
    # and ensures that none of the values in the list are None
    check_condition(
        isinstance(values, list) and all(value is not None for value in values),
        values,
        exception_message,
    )

    return values


def validate(
    value: any,
    type_of_validation: str,
    *,
    default_value: any = None,
    value_range: tuple[int, int] | None = None,
    exception_message: str = "Invalid value: {value}",
) -> any:
    """Validates the value according to the specified validation type

    Required arguments:
    :param value: (Any) the value to be validated
    :param type_of_validation: (String) the type of validation
        "boolean": checks if the value is of type bool
        "integer": checks if the value is of type int, and if the range of
        acceptable values is specified, verifies that the value is within the range
        "integer_list": checks if the value is of type list, and ensures that
        all values in the list are of type int. If the range of acceptable
        values is specified, verifies that each value in the list is within the range
        "any_list_wn": checks if the value is of type list, and ensures that
        none of the values in the list are None

    Optional configuration arguments:
    :param default_value: (Any) the default value. The function returns this value if it is False,
    but only if the validation type != "boolean"
    :param value_range: (Tuple[int, int]) the range of acceptable values.
    Should be set if the validation type = "integer"
    :param exception_message: (String) the error message if the value fails validation.
    Default value: "Invalid value: {value}"

    :return: the default value, in the case described above.
    If the validation is successful, returns the value that was initially
    passed to the function.
    """
    if type_of_validation != "boolean" and value is False:
        return default_value

    match type_of_validation:
        case "boolean":
            return boolean(value, exception_message)

        case "integer":
            return integer(value, exception_message, value_range)

        case "integer_list":
            return integer_list(value, exception_message, value_range)

        case "any_list_wn":
            return any_list_wn(value, exception_message)

        case _:
            raise InvalidValidationTypeError(validation_type=type_of_validation)



FULL_CHAR = "█"
HALF_CHAR = "▒"
TOTAL_PROGRESS = "[{current}/{total}]"  # [15/100]
PERCENT_PROGRESS = "[{percent:.2f}%]"  # [15.00%]
RAW_PROGRESS_BAR_LENGTH = 64


class ProgressBar:
    @staticmethod
    def show(current: int, total: int) -> None:
        percent = 100 * current / total
        current_bar_length = round(percent) // 2
        bar = current_bar_length * FULL_CHAR + (50 - current_bar_length) * HALF_CHAR
        print(
            bar,
            TOTAL_PROGRESS.format(current=current, total=total),
            PERCENT_PROGRESS.format(percent=percent),
            end="\r",
        )

    @staticmethod
    def get_length(total: int) -> int:
        return RAW_PROGRESS_BAR_LENGTH + len(str(total)) * 2



from dataclasses import dataclass

from ..utils import ConsoleColor
from .base import ApplicationException


class ValidatorException(ApplicationException):
    """General exception for validation errors"""

    @staticmethod
    def get_error_message(exception: "ValidatorException") -> None:
        print(
            ConsoleColor.paint_error("VALIDATION ERROR"),
            ConsoleColor.paint_info(exception.message),
            sep="\n",
        )


@dataclass(frozen=True, eq=False)
class InvalidValidationTypeError(ValidatorException):
    validation_type: str

    @property
    def message(self) -> str:
        return f"Invalid validation type: {self.validation_type}"


@dataclass(frozen=True, eq=False)
class InvalidValidationValueError(ValidatorException):
    value: str
    error_message: str

    @property
    def message(self) -> str:
        return self.error_message.format(value=self.value)



from dataclasses import dataclass, field

from ..constants import LAUNCH_TIME
from ..utils import call_counter
from .file_manager import FileManager


@dataclass
class YAMLOutputFile:
    data: dict[any, dict]
    folder_path: str = field(default="output")

    name: str = f"{LAUNCH_TIME.strftime("%d-%m-%Y-%H-%M-%S")}.yml"

    def __post_init__(self) -> None:
        self.file_path = FileManager.join_paths(self.folder_path, self.name)

    @call_counter
    def write_data(self, *data: any) -> None:
        """Writes data to the dictionary 'data'

        Example:
            write_data('Alexey', 'Yaroslav', 'TUParser')

        NOTE: the dictionary should have at least as many keys as the values passed.
        """
        for i, key in enumerate(self.data):
            self.data[key][self.write_data.calls] = data[i]

    def complete(self) -> None:
        FileManager.create_folder(self.folder_path)
        FileManager.dump_yaml(self.file_path, self.data)



from .base import ApplicationException
from .config import ConfigException, ConfigNotFoundError, InvalidConfigError
from .validator import InvalidValidationTypeError, InvalidValidationValueError, ValidatorException

__all__ = [
    "ApplicationException",
    "ConfigException",
    "ConfigNotFoundError",
    "InvalidConfigError",
    "ValidatorException",
    "InvalidValidationTypeError",
    "InvalidValidationValueError",
]



from .config import TelegraphParserConfig
from .constants import LAUNCH_TIME, TELEGRAPH_URL, TELEGRAPH_URL_MIRROR
from .file_handling import FileManager, YAMLOutputFile
from .parser import TelegraphParser, run_parser
from .utils import ConsoleColor, call_counter, get_monthrange, get_time_now
from .validator import validate

__all__ = [
    "TelegraphParserConfig",
    "LAUNCH_TIME",
    "TELEGRAPH_URL",
    "TELEGRAPH_URL_MIRROR",
    "FileManager",
    "YAMLOutputFile",
    "TelegraphParser",
    "run_parser",
    "ConsoleColor",
    "get_monthrange",
    "get_time_now",
    "call_counter",
    "validate",
]

# This information is used in the setup.py file:
__version__ = "0.1.5.5"
__author__ = "zombyacoff"
__email__ = "zombyacoff@gmail.com"
__git_url__ = "https://github.com/zombyacoff/tu-parser"



from calendar import monthrange
from datetime import datetime
from typing import Callable

from termcolor import colored


def get_time_now() -> datetime:
    return datetime.now()


def get_monthrange(month: int) -> int:
    """Returns the number of days in a given month"""
    return monthrange(2020, month)[1]


def call_counter(func: Callable) -> Callable:
    """Decorator that counts the number of times a function is called

    Usage:
        {FUNCTION_NAME}.calls

    Example:
        @call_counter
        def example_func():
            print(example_func.calls)

        example_func()  # 1
        example_func()  # 2
    """

    def wrapper(*args, **kwargs) -> int:
        wrapper.calls += 1
        return func(*args, **kwargs)

    wrapper.calls = 0
    return wrapper


class ConsoleColor:
    @staticmethod
    def paint_success(text: str) -> str:
        """Paints the given text in bold green"""
        return colored(text, "green", attrs=["bold"])

    @staticmethod
    def paint_error(text: str) -> str:
        """Paints the given text in bold red"""
        return colored(text, "red", attrs=["bold"])

    @staticmethod
    def paint_info(text: str) -> str:
        """Paints the given text in yellow"""
        return colored(text, "yellow")



from dataclasses import dataclass

from ..utils import ConsoleColor
from .base import ApplicationException


class ConfigException(ApplicationException):
    """General exception for configuration file errors"""

    @staticmethod
    def get_error_message(exception: "ConfigException") -> None:
        print(
            ConsoleColor.paint_error("CONFIGURATION FILE ERROR"),
            ConsoleColor.paint_info(exception.message),
            sep="\n",
        )


@dataclass(frozen=True, eq=False)
class ConfigNotFoundError(ConfigException):
    path: str

    @property
    def message(self) -> str:
        return f"The {self.path} YAML configuration file is missing"


@dataclass(frozen=True, eq=False)
class InvalidConfigError(ConfigException):
    error_message: str

    @property
    def message(self) -> str:
        return self.error_message



import asyncio
from abc import ABC, abstractmethod
from typing import Generator

import aiohttp
from bs4 import BeautifulSoup

from .config import TelegraphParserConfig
from .constants import LAUNCH_TIME, TELEGRAPH_URL
from .exceptions import ApplicationException
from .extensions import ProgressBar
from .file_handling import YAMLOutputFile
from .utils import ConsoleColor, call_counter, get_monthrange, get_time_now

SEMAPHORE_MAX_LIMIT = 150
HTTP_OK_STATUS = 200
PARSING_START_MESSAGE = """Parsing has started...
Do not turn off the program until the process is completed!\n"""
SUCCESS_COMPLETE_TITLE = "SUCCESSFULLY COMPLETED"
TIME_ELAPSED_TEXT = "Time elapsed: {time}"


class TelegraphParser(ABC):
    def __init__(self, config: TelegraphParserConfig) -> None:
        self.config = config

    async def __process_url(self, url: str) -> None:
        async with self.session.get(url) as page:
            if page.status != HTTP_OK_STATUS:
                return
            soup = BeautifulSoup(await page.text(), "html.parser")

        if not self.__check_release_date(soup):
            return

        await self.parse(url, soup)

    @abstractmethod
    async def parse(self, url: str, soup: BeautifulSoup) -> None: ...

    @call_counter
    def __get_progress_bar(self) -> None:
        if not self.config.progress_bar:
            return
        ProgressBar.show(self.__get_progress_bar.calls, self.config.total_urls)

    def __check_release_date(self, soup: BeautifulSoup) -> bool:
        if self.config.release_date is None:
            return True

        time_element = soup.select_one("time")
        release_date = (
            int(time_element.get_text("\n", strip=True)[-4:]) if time_element else LAUNCH_TIME.year
        )
        return release_date in self.config.release_date

    def get_complete_message(self) -> None:
        elapsed_time = get_time_now() - LAUNCH_TIME
        print(
            ConsoleColor.paint_success(SUCCESS_COMPLETE_TITLE)
            + " " * (ProgressBar.get_length(self.config.total_urls) - len(SUCCESS_COMPLETE_TITLE)),
            ConsoleColor.paint_info(TIME_ELAPSED_TEXT.format(time=elapsed_time)),
            sep="\n",
        )

    def __generate_urls(self) -> Generator[str, None, None]:
        for month in range(1, self.config.total_months + 1):
            for day in range(1, get_monthrange(month) + 1):
                for offset in range(1, self.config.offset + 1):
                    for title in self.config.titles:
                        url = f"{TELEGRAPH_URL}/{title}-{month:02}-{day:02}"
                        yield (f"{url}-{offset}" if offset > 1 else url)

    async def __semaphore_process(
        self,
        url: str,
        semaphore: asyncio.Semaphore,
    ) -> None:
        async with semaphore:
            await self.__process_url(url)

    async def main(self) -> None:
        print(ConsoleColor.paint_info(PARSING_START_MESSAGE))

        semaphore = asyncio.Semaphore(SEMAPHORE_MAX_LIMIT)
        urls_generator = self.__generate_urls()
        async with aiohttp.ClientSession() as self.session:
            processes = [self.__semaphore_process(url, semaphore) for url in urls_generator]
            for process in asyncio.as_completed(processes):
                await process
                self.__get_progress_bar()

        self.get_complete_message()

        # Complete 'output file' if child class
        # has an attribute whose type is YAMLOutputFile
        for _, attr_value in vars(self).items():
            if isinstance(attr_value, YAMLOutputFile):
                attr_value.complete()
                break


def run_parser(
    parser_class: TelegraphParser,
    *,
    config_class: TelegraphParserConfig = TelegraphParserConfig,
    parser_args: list[any] | None = None,
    config_path: str = "config",
) -> None:
    """Starts the parser

    Required arguments:
    :param parser_class: (TelegraphParser) the parser class,
    which must inherit from TelegraphParser

    Optional configuration arguments:
    :param config_class: (Config) custom configuration class, which must inherit from Config
    :param parser_args: (List) arguments passed to the constructor of the parser class
    :param config_path: (String) path to the YAML configuration file without extension. Default value: 'config'
    """
    try:
        config = config_class(config_path)
        parser = (
            parser_class(config) if parser_args is None else parser_class(config, *parser_args)
        )
        asyncio.run(parser.main())
    except ApplicationException as exception:
        exception.get_error_message(exception)



from textwrap import dedent

from setuptools import find_packages, setup

import tuparser


def get_requirements(file_name: str = "requirements.txt") -> list[str]:
    with open(file_name, "r") as file:
        return file.read().splitlines()


setup(
    name="tuparser",
    version=tuparser.__version__,
    author=tuparser.__author__,
    author_email=tuparser.__email__,
    url=tuparser.__git_url__,
    description=
    "A module that facilitates the creation of Telegraph parsing scripts",
    long_description=dedent(f"""\
        # Telegraph Universal Parser
        **Telegraph Universal Parser (tu-parser)** is a flexible module designed for creating custom parsers for the [Telegraph]({tuparser.TELEGRAPH_URL}) website.
        """),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=get_requirements(),
    python_requires=">=3.12",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="parser scraper parser-api telegraph telegraph-api",
)
